"""Top-level Stix Model Class."""
# standard library
import copy
import itertools
from operator import methodcaller
from typing import Dict, Iterable, Union

# third-party
import jmespath
from stix2.base import _STIXBase

# first-party
from tcex.batch import Batch
from tcex.logger import Logger


class StixModel:
    """STIX base model object."""

    def __init__(self, logger: Logger):
        """Initialize Class properties."""
        self.logger = logger
        self._as_object = None
        self._ipv4 = None
        self._ipv6 = None
        self._registry_key = None
        self._url = None
        self._domain_name = None
        self._email_address = None
        self._indicator = None
        self._relationship = None
        self._indicator_type_details = None
        self.default_map = {
            'dateAdded': '@.created',
            'lastModified': '@.modified',
            'tag': '@.labels',
            'securityLabel': '@.object_marking_refs',
            'attribute': [
                {'type': 'STIX Title', 'value': '@.name'},
                {'type': 'Description', 'value': '@.description'},
                {'type': 'STIX Indicator Type', 'value': '@.indicator_types'},
                {'type': 'External Date Created', 'value': '@.valid_from'},
                {'type': 'External Date Expires', 'value': '@.valid_until'},
                {'type': 'Source', 'value': 'WILL BE REPLACED IN THE CONSUME METHOD'},
            ],
        }

        self._visitors = []

    def _construct_summary(self, data, indicator_type):
        if 'summary' in data.keys():
            return data.get('summary')

        details = self.indicator_type_details.get(indicator_type)

        summary = []
        for field in details.get('fields', []):
            if field in data:
                summary.append(data.get(field))
            else:
                summary.append('')

        return ' : '.join(summary)

    @property
    def security_label_map(self):
        """Map of TLP labels to STIX marking-definitions."""
        return {
            'tlp:white': 'marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9',
            'tlp:green': 'marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da',
            'tlp:amber': 'marking-definition--f88d31f6-486f-44da-b317-01333bde0b82',
            'tlp:red': 'marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed',
        }

    @property
    def x_threat_rating_map(self):
        """Map of Threat Rating Numbers to Labels."""
        return {
            '0': 'Threat Rating: Unknown',
            '1': 'Threat Rating: Suspicious',
            '2': 'Threat Rating: Low',
            '3': 'Threat Rating: Moderate',
            '4': 'Threat Rating: High',
            '5': 'Threat Rating: Very High',
        }

    @property
    def indicator_type_details(self):
        """Details for individual ThreatConnect indicator types."""
        if not self._indicator_type_details:
            # first-party
            from tcex.stix.indicator.stix_pattern_helpers import (
                address_stix_pattern_producer,
                asn_stix_pattern_producer,
                cidr_stix_pattern_producer,
                email_address_stix_pattern_producer,
                file_stix_pattern_producer,
                host_stix_pattern_producer,
                registery_key_stix_pattern_producer,
                url_stix_pattern_producer,
            )

            self._indicator_type_details = {
                'host': {
                    'lambda': host_stix_pattern_producer,
                    'api_branch': 'host',
                    'fields': ['text'],
                },
                'url': {
                    'lambda': url_stix_pattern_producer,
                    'api_branch': 'url',
                    'fields': ['text'],
                },
                'emailaddress': {
                    'lambda': email_address_stix_pattern_producer,
                    'api_branch': 'emailaddress',
                    'fields': ['addresses'],
                },
                'asn': {
                    'lambda': asn_stix_pattern_producer,
                    'api_branch': 'asns',
                    'fields': ['as_number'],
                },
                'address': {
                    'lambda': address_stix_pattern_producer,
                    'api_branch': 'address',
                    'fields': ['ip'],
                },
                'cidr': {
                    'lambda': cidr_stix_pattern_producer,
                    'api_branch': 'cidr',
                    'fields': ['ip'],
                },
                'file': {
                    'lambda': file_stix_pattern_producer,
                    'api_branch': 'file',
                    'fields': ['md5', 'sha1', 'sha256'],
                },
                'registry key': {
                    'lambda': registery_key_stix_pattern_producer,
                    'api_branch': 'registryKey',
                    'fields': ['Key Name', 'Value Name', 'Value Type'],
                },
            }
        return self._indicator_type_details

    @property
    def as_object(self):
        """ASN Parser."""
        if not self._as_object:
            from .observables.autonomous_system import StixASObject

            self._as_object = StixASObject(self.logger)
        return self._as_object

    @property
    def ipv4(self):
        """IPv4 Parser."""
        if not self._ipv4:
            from .observables.ip_addr import StixIPv4Object

            self._ipv4 = StixIPv4Object(self.logger)
        return self._ipv4

    @property
    def ipv6(self):
        """IPv6 Parser."""
        if not self._ipv6:
            from .observables.ip_addr import StixIPv6Object

            self._ipv6 = StixIPv6Object(self.logger)
        return self._ipv6

    @property
    def registry_key(self):
        """Registry Key Parser."""
        if not self._registry_key:
            from .observables.registry_key import StixRegistryKeyObject

            self._registry_key = StixRegistryKeyObject(self.logger)
        return self._registry_key

    @property
    def url(self):
        """URL Parser."""
        if not self._url:
            from .observables.url import StixURLObject

            self._url = StixURLObject(self.logger)
        return self._url

    @property
    def email_address(self):
        """Email Address Parser."""
        if not self._email_address:
            from .observables.email_address import StixEmailAddressObject

            self._email_address = StixEmailAddressObject(self.logger)
        return self._email_address

    @property
    def domain_name(self):
        """Domain Name Parser."""
        if not self._domain_name:
            from .observables.domain_name import StixDomainNameObject

            self._domain_name = StixDomainNameObject(self.logger)
        return self._domain_name

    @property
    def indicator(self):
        """Return Indicator Parser."""
        if not self._indicator:
            from .indicator.indicator import StixIndicator

            self._indicator = StixIndicator(self.logger)
        return self._indicator

    @property
    def relationship(self):
        """Relationship Parser."""
        if not self._relationship:
            from .relationship.relationship import Relationship

            self._relationship = Relationship()
        return self._relationship

    def produce(
        self, tc_data: Union[list, dict], type_mapping: Dict = None, **kwargs
    ):  # pylint: disable=unused-argument
        """Convert ThreatConnect data (in parsed JSON format) into STIX objects.

        Args:
            tc_data: one or more ThreatConnect object dictionaries
            type_mapping: mapping of TC type to a StixModel object that can produce() it.

        Yields:
            STIX objects
        """

        type_mapping = type_mapping or {}

        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            for indicator_type, normalized_data in self._normalize_tc_objects(data):
                if indicator_type in type_mapping:
                    yield from type_mapping.get(indicator_type).produce(
                        normalized_data, indicator_type=indicator_type
                    )
                else:
                    yield from self.relationship.produce(normalized_data)
                    yield from self.indicator.produce(
                        normalized_data, indicator_type=indicator_type
                    )

    def _normalize_tc_objects(self, tc_data):
        indicator_type = None
        api_branch = 'indicator'
        data = tc_data.get('data', tc_data)
        for key, values in self.indicator_type_details.items():
            if values.get('api_branch').upper() in [s.upper() for s in data.keys()] or values.get(
                'type', ''
            ).upper() in [s.upper() for s in data.keys()]:
                api_branch = values.get('api_branch')
                indicator_type = key
                break

        if indicator_type is None:
            for key, values in self.indicator_type_details.items():
                for field in values.get('fields'):
                    if field.lower() in map(str.lower, data.keys()):
                        indicator_type = key
                        break
                if indicator_type:
                    break

        tc_data = data.get(api_branch, data)

        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            _type = indicator_type or data.get('type').lower()
            if _type not in self.indicator_type_details:
                self.logger.log.info(f'unable to convert indicator type {_type} to stix object')
                continue
            data['summary'] = self._construct_summary(data, _type)
            for field in self.indicator_type_details.get(_type).get('fields'):
                data.pop(field, '')
            yield _type, data

    # pylint: disable=unused-argument,no-self-use
    def as_object_mapping(self, stix_data):
        """Produce ThreatConnect ASN mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        return {
            'type': 'ASN',
            'summary': '@.number',
            'confidence': '@.confidence',
        }

    # pylint: disable=unused-argument,no-self-use
    def domain_name_mapping(self, stix_data):
        """Produce ThreatConnect Host mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        return {
            'type': 'Host',
            'summary': '@.value',
            'confidence': '@.confidence',
        }

    # pylint: disable=unused-argument,no-self-use
    def email_address_mapping(self, stix_data):
        """Produce ThreatConnect EmailAddress mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        return {
            'type': 'EmailAddress',
            'summary': '@.value',
            'confidence': '@.confidence',
        }

    # pylint: disable=unused-argument,no-self-use
    def _ip_addr_mapping(self, stix_data, full_block_size):
        """Produce ThreatConnect Address/CIDR mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        cidr_parts = stix_data.get('value', '').split('/')
        cidr_suffix = cidr_parts[1] if len(cidr_parts) > 1 else str(full_block_size)
        if cidr_suffix == str(full_block_size):
            return {
                'type': 'Address',
                'summary': '@.value',
                'confidence': '@.confidence',
            }
        return {
            'confidence': '@.confidence',
            'type': 'CIDR',
            'summary': '@.value',
        }

    # pylint: disable=unused-argument,no-self-use
    def ipv6_mapping(self, stix_data):
        """Produce ThreatConnect Address/CIDR mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        return self._ip_addr_mapping(stix_data, 128)

    # pylint: disable=unused-argument,no-self-use
    def ipv4_mapping(self, stix_data):
        """Produce ThreatConnect Address/CIDR mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        return self._ip_addr_mapping(stix_data, 32)

    # pylint: disable=unused-argument,no-self-use
    def url_mapping(self, stix_data):
        """Produce ThreatConnect URL mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        return {'type': 'URL', 'summary': '@.value', 'confidence': '@.confidence'}

    # pylint: disable=unused-argument,no-self-use
    def registry_key_mapping(self, stix_data):
        """Produce ThreatConnect Registry Key mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A indicator mappings.
        """

        mapper = {
            'type': 'Registry Key',
            'Key Name': '@.key',
            'confidence': '@.confidence',
        }
        if not stix_data.get('values'):
            return mapper

        for i in range(len(stix_data.get('values'))):
            mapper['Value Name'] = f'@.values[{i}].name'
            mapper['Value Type'] = f'@.values[{i}].data_type'
            mapper.setdefault('attribute', []).append(
                {'type': 'Value Data', 'value': f'@.values[{i}].data'}
            )
        return mapper

    def consume(
        self,
        stix_data: dict,
        collection_id,
        collection_name,
        collection_path,
        custom_type_mapping: dict = None,
    ):
        """Convert stix_data (in parsed JSON format) into ThreatConnect objects.

        Args:
            stix_data: one or more stix_data dictionaries
            collection_path: the url path to the collection
            collection_name: the collection name
            collection_id: the collection id
            custom_type_mapping: stix type to a mapping function which takes stix_data as a param.

        Yields:
            ThreatConnect objects
        """
        type_mapping = {
            'email-addr': self.email_address_mapping,
            'autonomous-system': self.as_object_mapping,
            'domain-name': self.domain_name_mapping,
            'ipv4-addr': self.ipv4_mapping,
            'ipv6-addr': self.ipv6_mapping,
            'windows-registry-key': self.registry_key_mapping,
            'url': self.url_mapping,
        }
        type_mapping.update(custom_type_mapping or {})
        visitor_mapping = {'relationship': self.relationship}

        tc_data = []
        for data in stix_data.get('objects', []):
            object_id = data.get('id')
            xid = Batch.generate_xid(object_id)
            safe_default_map = copy.deepcopy(self.default_map)
            source_value = [f'Object ID: {object_id}']
            if collection_id:
                source_value.append(f'Collection ID: {collection_id}')
                xid = Batch.generate_xid([object_id, collection_id])
            if collection_name:
                source_value.append(f'Collection Name: {collection_name}')
            if collection_path:
                source_value.append(f'Collection Path: {collection_path}')
                if not collection_path.endswith('/'):
                    collection_path += '/'
                source_value.append(f'Object Path: {collection_path}objects/{object_id}/')
            for attribute in safe_default_map.get('attribute', []):
                if attribute.get('type') == 'Source':
                    attribute['value'] = '\n'.join(source_value)
                    attribute['displayed'] = True
                    break
            _type = data.get('type').lower()
            mapping_method = visitor_mapping.get(_type)
            if mapping_method:
                # for visitor in mapping_method.consume(data):
                #     self.register_visitor(visitor)
                continue
            mapping_method = type_mapping.get(_type)
            if _type == 'indicator':
                if stix_data.get('pattern_type', 'stix') != 'stix':
                    continue
                for map_ in self.indicator.consume_mappings(data, collection_id):
                    if not map_:
                        continue
                    map_.update(safe_default_map)
                    tc_data = itertools.chain(tc_data, self._map(data, map_))
            elif mapping_method:
                safe_default_map['xid'] = xid
                map_ = self.smart_update(safe_default_map, mapping_method(data))
                tc_data = itertools.chain(tc_data, self._map(data, map_))
            # else:
            #     self.default_map['xid'] = xid
            #     tc_data = itertools.chain(tc_data, self._map(data, self.default_map))
            #     continue

        for visitor in self._visitors:
            tc_data = visitor.visit(tc_data)

        yield from tc_data

    @staticmethod
    def smart_update(list_one, list_two):
        """Update one dict with the other but joins arrays."""
        updated_dict = {}
        dict_items = map(methodcaller('items'), (list_one, list_two))
        for k, v in itertools.chain.from_iterable(dict_items):
            if isinstance(v, list) and isinstance(updated_dict.get(k), list):
                updated_dict[k].extend(v)
            else:
                updated_dict[k] = v
        return updated_dict

    def register_visitor(self, visitor):
        """Register a visitor that will be passed all parsed data after consume is through."""
        self._visitors.append(visitor)

    def _map(self, data: Union[list, dict], mapping: dict):
        """Produce a dict with the appropriate values given data and a mapping.

        Args:
            data: The data to be referenced in the mappings
            mapping: The default values and jmspaths to the appropiate values in data.

        Returns:
            A new dict produced by the mappings and data field.
        """
        if isinstance(data, dict):
            data = [data]
        for d in data:
            mapped_obj = mapping.copy()
            for key, value in mapping.items():
                if isinstance(value, str) and not value.startswith('@'):
                    mapped_obj[key] = value
                    continue
                if isinstance(value, (dict, list)):
                    value = copy.deepcopy(value)

                if key in ['securityLabel', 'tag', 'attribute']:
                    mapped_value = self._custom_stix_mapping(d, key, value)
                    if mapped_value:
                        mapped_obj[key] = mapped_value
                    else:
                        del mapped_obj[key]
                else:
                    if isinstance(value, list):
                        new_list = []
                        for item in value:
                            new_list.append(list(self._map(d, item))[0])

                        mapped_obj[key] = new_list
                        continue
                    if isinstance(value, dict):
                        mapped_obj[key] = list(self._map(d, mapped_obj[key]))[0]
                        continue

                    resolved_value = jmespath.search(f'{value}', jmespath.search('@', d)) or []
                    if resolved_value:
                        mapped_obj[key] = resolved_value
                    else:
                        del mapped_obj[key]
            yield mapped_obj

    @staticmethod
    def _remove_milliseconds(time):
        time = time.split('.')
        milliseconds = ''
        if len(time) > 1:
            milliseconds = time.pop()
        new_time = ''.join(time)
        if milliseconds.lower().endswith('z'):
            return f'{new_time}Z'
        return new_time

    def _custom_stix_mapping(self, data, key, value):
        if key == 'securityLabel':
            resolved_values = jmespath.search(f'{value}', jmespath.search('@', data)) or []
            object_marking_refs = []
            for resolved_value in resolved_values:
                resolved_value = resolved_value.lower()
                if resolved_value in self.security_label_map.values():
                    security_label = list(self.security_label_map.keys())[
                        list(self.security_label_map.values()).index(resolved_value)
                    ]
                    object_marking_refs.append({'name': security_label.upper()})
            return object_marking_refs
        if key == 'tag':
            resolved_values = jmespath.search(f'{value}', jmespath.search('@', data)) or []
            tags = []
            for resolved_value in resolved_values:
                tags.append({'name': resolved_value})
            return tags
        if key == 'attribute':
            attributes = []
            for attribute in value:
                attribute_value = attribute.get('value')
                if isinstance(attribute_value, str):
                    if attribute.get('value', '').startswith('@'):
                        attribute_value = (
                            jmespath.search(f'{attribute_value}', jmespath.search('@', data)) or []
                        )
                        if isinstance(attribute_value, list):
                            attribute_value = '\n'.join(attribute_value)
                        if attribute.get('value') in ['@.valid_from', '@.valid_until']:
                            attribute_value = self._remove_milliseconds(attribute_value)
                if not attribute_value:
                    continue
                attribute['value'] = attribute_value
                attributes.append(attribute)
            return attributes
        return None

        # return mapped_obj
        #     if not object_marking_refs:
        #         del mapped_obj[key]
        #     else:
        #         mapped_obj[key] = []
        #     for object_marking_ref in object_marking_refs:
        #         object_marking_ref = object_marking_ref.lower()
        #         if object_marking_ref in self.security_label_map.values():
        #             security_label = list(self.security_label_map.keys())[
        #                 list(self.security_label_map.values()).index(object_marking_ref)]
        #             mapped_obj[key].append({'name': security_label.upper()})
        # elif key == 'tag':
        #     tags = jmespath.search(f'{value}', jmespath.search('@', d)) or []
        #     if not tags:
        #         del mapped_obj[key]
        #     else:
        #         mapped_obj = []
        #     for tag in tags:
        #         mapped_obj[key].append({'name': tag})
        # else:
        #     if key == 'attribute':
        #         attributes = jmespath.search(f'{value}', jmespath.search('@', d)) or []
        #         if not attributes:
        #             del mapped_obj[key]
        #     mapped_obj[key] = jmespath.search(f'{value}', jmespath.search('@', d))


class JMESPathStixModel(StixModel):
    """Generic implmenetation of StixModel that converts data based on jmespath statements."""

    def __init__(
        self,
        produce_map: Dict[str, str],
        produce_type: _STIXBase,
        consume_map: Dict[str, str],
        logger: Logger,
    ):
        """Instantiate a JMESPath StixModel.

        Args:
            produce_map: map of target field name to a jmespath query to pull that data from the
                tc object.
            produce_type: The class to use for the STIX object.
            consume_map: map of target field name to a jmespath query to pull data from the stix
                object.
            logger: logger
        """
        super().__init__(logger)
        self._produce_map = produce_map
        self._produce_type = produce_type
        self._consume_map = consume_map

    # pylint: disable=arguments-differ
    def produce(self, tc_data: Union[list, dict], **kwargs) -> Iterable:
        """Accept TC entities and produces STIX objects."""
        yield from (
            self._produce_type(**stix_data) for stix_data in self._map(tc_data, self._produce_map)
        )

    # pylint: disable=arguments-differ, unused-argument
    def consume(self, stix_data: Union[list, dict], **kwargs):
        """Accept STIX objects and produces TC Entities."""
        yield from self._map(stix_data, self._consume_map)



"""Parser for STIX Indicator Objects.

import stix2
see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070633
"""
# standard library
import itertools
import uuid
from datetime import datetime
from typing import Iterable, Union

# first-party
from tcex.batch import Batch
from tcex.stix import StixModel  # pylint: disable=cyclic-import

from .stix2_lark import Indicator, Stix2IndicatorParser


class StixIndicator(StixModel):
    """Parser for STIX Indicator Objects.

    see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070633
    """

    def __init__(self, logger):
        """Initialize stix parser."""
        super().__init__(logger)
        self.stix_parser = Stix2IndicatorParser()

    @staticmethod
    def _add_milliseconds(time):
        new_time = f'''{time.upper().replace('Z', '')}.000'''
        if time.lower().endswith('z'):
            return f'{new_time}Z'
        return new_time

    # pylint: disable=arguments-differ
    def produce(self, tc_data: Union[list, dict], **kwargs):
        """Produce a STIX Indicator from a ThreatConnect Indicator.

        Args:
            tc_data: ThreatConnect indicator(s)

        Yields:
            A STIX Indicator.
        """
        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            _type = kwargs.get('indicator_type') or data.get('type')
            indicator_details = self.indicator_type_details.get(_type.lower())
            if not indicator_details:
                continue
            kwargs = {
                'pattern_type': 'stix',
                'lang': 'en',
                'pattern_version': '2.1',
                'type': 'indicator',
                'created': self._add_milliseconds(data.get('dateAdded')),
                'valid_from': self._add_milliseconds(data.get('dateAdded')),
                'modified': self._add_milliseconds(data.get('lastModified')),
                'name': data.get('summary'),
                'pattern': indicator_details.get('lambda')(data),
                'description': '',
                'indicator_types': ['malicious-activity'],
                'confidence': data.get('confidence', 0),
            }
            if not data.get('active', True):
                kwargs['revoked'] = True
            latest = None
            for tag in data.get('tag', []):
                kwargs.setdefault('labels', []).append(tag.get('name'))
            for security_label in data.get('securityLabel', []):
                security_label = security_label.get('name', '').strip().lower()
                if security_label in self.security_label_map:
                    kwargs.setdefault('object_marking_refs', [])
                    kwargs['object_marking_refs'].append(
                        self.security_label_map.get(security_label)
                    )
            for attribute in data.get('attribute', []):
                if attribute.get('type').lower() == 'description':
                    value = attribute.get('value')
                    last_modified = datetime.strptime(
                        attribute.get('lastModified'), '%Y-%m-%dT%H:%M:%SZ'
                    )
                    if attribute.get('displayed'):
                        kwargs['description'] = value
                        break
                    if not latest or last_modified > latest:
                        latest = last_modified
                        kwargs['description'] = value

            # APP-1766 - STIX spec requirement
            id_ = f'''{data.get('ownerName').lower()}-{_type.lower()}-{data.get('summary')}'''
            if id_.lower() == '00abedb4-aa42-466c-9c01-fed23315a9b7':
                self.logger.log.error(
                    'RESERVED UUID 00abedb4-aa42-466c-9c01-fed23315a9b7 '
                    f'created for indicator {data}'
                )
                continue
            kwargs['id'] = f'indicator--{uuid.uuid5(uuid.NAMESPACE_X500, id_)}'

            # {
            #     "type": "indicator",
            #     "spec_version": "2.1",
            #     "id": "indicator--78e53a5c-9510-4a95-88b1-dbac1ee60ca5",
            #     "created": "2020-09-08T19:16:25.481964Z",
            #     "modified": "2020-09-08T19:16:25.481964Z",
            #     "name": "TCI - kqvri.com",
            #     "description": "visible description",
            #     "indicator_types": [
            #         "malicious-activity"
            #     ],
            #     "pattern": "[domain-name:value = 'kqvri.com']",
            #     "pattern_type": "stix",
            #     "pattern_version": "2.1",
            #     "valid_from": "2020-09-08T19:16:25.481964Z",
            #     "labels": [
            #         "test1"
            #     ]
            # }

            try:
                rating = str(int(data.get('rating')))
                kwargs.setdefault('labels', []).append(self.x_threat_rating_map[rating])
            except Exception:  # nosec
                pass

            yield kwargs

    def consume_mappings(self, stix_data: dict, collection_id: str):
        """Produce ThreatConnect mappings from a STIX 2.1 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A array of indicator mappings.
        """
        mappings = []
        batch_xid_array = [stix_data.get('id')]
        if collection_id:
            batch_xid_array += [collection_id]
        try:
            indicators = self.stix_parser.parse(stix_data.get('pattern'))
            filtered_indicators = []
            for indicator in indicators:
                if not hasattr(indicator, 'path'):
                    self.logger.log.info(
                        '''Pattern not supported. '''
                        f'''Ignoring tree indicators for id: {stix_data.get('id')}.'''
                    )
                    continue
                filtered_indicators.append(indicator)
            indicators = filtered_indicators
            mappings = [
                self._default_consume_handler(indicators, batch_xid_array),
                self._ip_consume_handler(indicators, batch_xid_array),
                self._file_consume_handler(indicators, batch_xid_array),
            ]
            mappings = list(itertools.chain(*mappings))
        except Exception:
            self.logger.log.trace(f'''Error occurred parsing pattern: {stix_data.get('pattern')}''')
        return mappings

    @staticmethod
    def _file_consume_handler(indicators: Iterable[Indicator], batch_xid_array):
        """Produce ThreatConnect file mappings from a list of STIX 2.1 indicators

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A array of indicator mappings.
        """
        file_indicators = list(filter(lambda i: 'file:hashes' in i.path, indicators))

        sha256_indicators = list(filter(lambda i: 'SHA-156' in i.path.upper(), file_indicators))
        sha2_indicators = list(filter(lambda i: 'SHA-1' in i.path.upper(), file_indicators))
        md5_indicators = list(filter(lambda i: 'MD5' in i.path.upper(), file_indicators))

        if len(file_indicators) <= 0:
            return []
        batch_xid_array.append('File')
        mappings = []
        if (
            len(file_indicators) <= 3
            and len(sha256_indicators) <= 1
            and len(sha2_indicators) <= 1
            and len(md5_indicators) <= 1
        ):
            value = ' : '.join([v.value for v in file_indicators])
            mappings.append(
                {
                    'type': 'File',
                    'summary': value,
                    'confidence': '@.confidence',
                    'xid': Batch.generate_xid(batch_xid_array + [value]),
                }
            )
        else:
            for i in file_indicators:
                mappings.append(
                    {
                        'type': 'File',
                        'summary': i.value,
                        'confidence': '@.confidence',
                        'xid': Batch.generate_xid(batch_xid_array + [i.value]),
                    }
                )
        return mappings

    @staticmethod
    def _ip_consume_handler(indicators: Iterable[Indicator], batch_xid_array):
        """Produce ThreatConnect Address/CIDR mappings from a list of STIX 2.1 indicators

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A array of indicator mappings.
        """
        mappings = []
        for i in filter(lambda i: i.path in ['ipv4-addr:value', 'ipv6-addr:value'], indicators):
            parse_map = None
            if i.path == 'ipv4-addr:value':
                if '/' in i.value and i.value.split('/')[1] != '32':  # this is a CIDR
                    parse_map = {
                        'type': 'CIDR',
                        'summary': i.value,
                    }
                else:  # this is an address
                    parse_map = {
                        'type': 'Address',
                        'summary': i.value.split('/')[0],
                    }
            elif i.path == 'ipv6-addr:value':
                if '/' in i.value and i.value.split('/')[1] != '128':  # this is a CIDR
                    parse_map = {
                        'type': 'CIDR',
                        'summary': i.value,
                    }
                else:  # this is an address
                    parse_map = {
                        'type': 'Address',
                        'summary': i.value.split('/')[0],
                    }
            parse_map['confidence'] = '@.confidence'
            parse_map['xid'] = Batch.generate_xid(
                batch_xid_array + [parse_map.get('type'), parse_map.get('summary')]
            )
            mappings.append(parse_map)
        return mappings

    @staticmethod
    def _default_consume_handler(indicators: Iterable[Indicator], batch_xid_array):
        """Produce ThreatConnect URL/EmailAddress/Host/ASN mappings from STIX 2.1 indicators

        Args:
            stix_data: STIX Indicator objects to parse.

        Returns:
            A array of indicator mappings.
        """

        type_map = {
            'url:value': 'URL',
            'email-addr:value': 'EmailAddress',
            'domain-name:value': 'Host',
            'autonomous-system:name': 'ASN',
        }

        mappings = []
        for i in filter(lambda i: i.path in type_map.keys(), indicators):
            indicator_type = type_map.get(i.path)

            mappings.append(
                {
                    'type': indicator_type,
                    'summary': i.value,
                    'confidence': '@.confidence',
                    'xid': Batch.generate_xid(batch_xid_array + [indicator_type, i.value]),
                }
            )
        return mappings



"""TcEx Framework Key Value API Module"""
# standard library
from typing import Any
from urllib.parse import quote


class KeyValueApi:
    """TcEx Key Value API Module.

    Args:
        session (request.Session): A configured requests session for TC API (tcex.session).
        runtime_level: The runtime level of the App.
    """

    def __init__(self, session: object, runtime_level: str):
        """Initialize the Class properties."""
        self._runtime_level = runtime_level
        self._session = session

    def create(self, context: str, key: str, value: Any) -> str:
        """Create key/value pair in remote KV store.

        Args:
            context: A specific context for the create.
            key: The key to create in remote KV store.
            value: The value to store in remote KV store.

        Returns:
            (string): The response from the API call.
        """
        key: str = quote(key, safe='~')
        headers = {'content-type': 'application/octet-stream'}

        # this conditional is only required while there are TC instances < 6.0.7 in the wild.
        # once all TC instance are > 6.0.7 the context endpoint should work for PB Apps.
        url = f'/internal/playbooks/keyValue/{key}'
        if self._runtime_level in ['apiservice', 'triggerservice', 'webhooktriggerservice']:
            url = f'/internal/playbooks/keyValue/{context}/{key}'
        r = self._session.put(url, data=value, headers=headers)
        return r.content

    def read(self, context: str, key: str) -> Any:
        """Read data from remote KV store for the provided key.

        Args:
            context: A specific context for the create.
            key: The key to read in remote KV store.

        Returns:
            (any): The response data from the remote KV store.
        """
        key = quote(key, safe='~')

        # this conditional is only required while there are TC instances < 6.0.7 in the wild.
        # once all TC instance are > 6.0.7 the context endpoint should work for PB Apps.
        url = f'/internal/playbooks/keyValue/{key}'
        if self._runtime_level in ['apiservice', 'triggerservice', 'webhooktriggerservice']:
            url = f'/internal/playbooks/keyValue/{context}/{key}'
        r = self._session.get(url)
        data = r.content

        # Binary data for PB Apps is base64 encoded, for service Apps it is not
        if data is not None and isinstance(data, bytes):
            data = data.decode('utf-8')
        return data



"""TcEx Framework Key Value Redis Module"""
# standard library
from typing import Any, Optional


class KeyValueRedis:
    """TcEx Key Value Redis Module.

    Args:
        redis_client (redis.Client): An instance of redis client.
    """

    def __init__(self, redis_client: object):
        """Initialize the Class properties."""
        self._redis_client = redis_client

    def create(self, context: str, key: str, value: Any) -> None:
        """Create key/value pair in Redis.

        Args:
            context: A specific context for the create.
            key (str): The field name (key) for the kv pair in Redis.
            value (any): The value for the kv pair in Redis.

        Returns:
            str: The response from Redis.
        """
        return self._redis_client.hset(context, key, value)

    def delete(self, context: str, key: str) -> str:
        """Alias for hdel method.

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.

        Returns:
            str: The response from Redis.
        """
        return self._redis_client.hdel(context, key)

    def hgetall(self, context: str):
        """Read data from Redis for the current context.

        Args:
            context: A specific context for the create.

        Returns:
            list: The response data from Redis.
        """
        return self._redis_client.hgetall(context)

    def read(self, context: str, key: str, decode='utf-8') -> Any:
        """Read data from Redis for the provided key.

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.
            decode: encoding to use to decode retrieved value or False to not decode value.

        Returns:
            str: The response data from Redis.
        """
        value = self.hget(context, key)
        # convert retrieved bytes to string
        if isinstance(value, bytes) and decode:
            value = value.decode('utf-8')
        return value

    def hget(self, context: str, key: str) -> Optional[bytes]:
        """Read data from redis for the provided key.

        This method will *not* convert the retrieved data (like read() does).

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.

        Returns:
            Optional[bytes]: the raw value from redis, if any
        """
        return self._redis_client.hget(context, key)



"""ThreatConnect Batch Import Module."""
# standard library
import gzip
import hashlib
import json
import os
import re
import shelve
import sys
import time
import uuid
from collections import deque
from typing import Optional, Tuple, Union

from .group import (
    Adversary,
    Campaign,
    Document,
    Email,
    Event,
    Group,
    Incident,
    IntrusionSet,
    Report,
    Signature,
    Threat,
)
from .indicator import (
    ASN,
    CIDR,
    URL,
    Address,
    EmailAddress,
    File,
    Host,
    Indicator,
    Mutex,
    RegistryKey,
    UserAgent,
    custom_indicator_class_factory,
)

# import local modules for dynamic reference
module = __import__(__name__)


class BatchWriter:
    """ThreatConnect Batch Import Module"""

    def __init__(self, tcex: 'TcEx', output_dir: str, **kwargs) -> None:  # noqa: F821
        """Initialize Class properties."""
        self.tcex = tcex
        self.output_dir = output_dir
        self.output_extension = kwargs.get('output_extension')
        self.write_callback = kwargs.get('write_callback')
        self.write_callback_kwargs = kwargs.get('write_callback_kwargs', {})

        # properties
        self._batch_files = []
        self._batch_max_chunk = 100_000
        self._batch_size = 0  # track current batch size
        self._batch_max_size = 75_000_000  # max size in bytes
        # self._batch_max_size = 7_500_000  # max size in bytes

        # shelf settings
        self._group_shelf_fqfn = None
        self._indicator_shelf_fqfn = None

        # containers
        self._groups = None
        self._groups_shelf = None
        self._indicators = None
        self._indicators_shelf = None

        # build custom indicator classes
        self._gen_indicator_class()

    def _gen_indicator_class(self):  # pragma: no cover
        """Generate Custom Indicator Classes."""
        for entry in self.tcex.indicator_types_data.values():
            name = entry.get('name')
            class_name = name.replace(' ', '')
            # temp fix for API issue where boolean are returned as strings
            entry['custom'] = self.tcex.utils.to_bool(entry.get('custom'))

            if class_name in globals():
                # skip Indicator Type if a class already exists
                continue

            # Custom Indicator can have 3 values. Only add the value if it is set.
            value_fields = []
            if entry.get('value1Label'):
                value_fields.append(entry['value1Label'])
            if entry.get('value2Label'):
                value_fields.append(entry['value2Label'])
            if entry.get('value3Label'):
                value_fields.append(entry['value3Label'])
            value_count = len(value_fields)

            class_data = {}
            # Add Class for each Custom Indicator type to this module
            custom_class = custom_indicator_class_factory(name, Indicator, class_data, value_fields)
            setattr(module, class_name, custom_class)

            # Add Custom Indicator Method
            self._gen_indicator_method(name, custom_class, value_count)

    def _gen_indicator_method(
        self, name: str, custom_class: object, value_count: int
    ) -> None:  # pragma: no cover
        """Dynamically generate custom Indicator methods.

        Args:
            name (str): The name of the method.
            custom_class (object): The class to add.
            value_count (int): The number of value parameters to support.
        """
        method_name = name.replace(' ', '_').lower()

        # Add Method for each Custom Indicator class
        def method_1(value1: str, xid, **kwargs):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            indicator_obj = custom_class(value1, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        def method_2(
            value1: str, value2: str, xid, **kwargs
        ):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            indicator_obj = custom_class(value1, value2, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        def method_3(
            value1: str, value2: str, value3: str, xid, **kwargs
        ):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            indicator_obj = custom_class(value1, value2, value3, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        method = locals()[f'method_{value_count}']
        setattr(self, method_name, method)

    def _group(
        self, group_data: Union[dict, object], store: Optional[bool] = True
    ) -> Union[dict, object]:
        """Return previously stored group or new group.

        Args:
            group_data: An Group dict or instance of Group object.
            store: If True the group data will be stored in instance list.

        Returns:
            Union[dict, object]: The new Group dict/object or the previously stored dict/object.
        """
        if store is False:
            return group_data

        if isinstance(group_data, dict):
            # get xid from dict
            xid = group_data.get('xid')
        else:
            # get xid from object
            xid = group_data.xid

        if self.groups.get(xid) is not None:
            # return existing group from memory
            group_data = self.groups.get(xid)
        elif self.groups_shelf.get(xid) is not None:
            # return existing group from shelf
            group_data = self.groups_shelf.get(xid)
        else:
            # store new group
            self.groups[xid] = group_data

            # track total batch job data size as TI gets added
            if isinstance(group_data, dict):
                self._batch_size += sys.getsizeof(json.dumps(group_data))
            else:
                self._batch_size += sys.getsizeof(json.dumps(group_data.data))

            # max size hit, dump TI to disk
            if self._batch_size > self._batch_max_size:
                self.dump()
        return group_data

    def _indicator(
        self, indicator_data: Union[dict, object], store: Optional[bool] = True
    ) -> Union[dict, object]:
        """Return previously stored indicator or new indicator.

        Args:
            indicator_data: An Indicator dict or instance of Indicator object.
            store: If True the indicator data will be stored in instance list.

        Returns:
            Union[dict, object]: The new Indicator dict/object or the previously stored dict/object.
        """
        if store is False:
            return indicator_data

        if isinstance(indicator_data, dict):
            # get xid from dict
            xid = indicator_data.get('xid')
        else:
            # get xid from object
            xid = indicator_data.xid

        if self.indicators.get(xid) is not None:
            # return existing indicator from memory
            indicator_data = self.indicators.get(xid)
        elif self.indicators_shelf.get(xid) is not None:
            # return existing indicator from shelf
            indicator_data = self.indicators_shelf.get(xid)
        else:
            # store new indicators
            self.indicators[xid] = indicator_data

            # track total batch job data size as TI gets added
            if isinstance(indicator_data, dict):
                self._batch_size += sys.getsizeof(json.dumps(indicator_data))
            else:
                self._batch_size += sys.getsizeof(json.dumps(indicator_data.data))

            # max size hit, dump TI to disk
            if self._batch_size > self._batch_max_size:
                self.dump()
        return indicator_data

    @staticmethod
    def _indicator_values(indicator: str) -> list:
        """Process indicators expanding file hashes/custom indicators into multiple entries.

        Args:
            indicator: Indicator value represented as " : " delimited string.

        Returns:
            list: The list of indicators split on " : ".
        """
        indicator_list = [indicator]
        if indicator.count(' : ') > 0:
            # handle all multi-valued indicators types (file hashes and custom indicators)
            indicator_list = []

            # group 1 - lazy capture everything to first <space>:<space> or end of line
            iregx_pattern = r'^(.*?(?=\s\:\s|$))?'
            iregx_pattern += r'(?:\s\:\s)?'  # remove <space>:<space>
            # group 2 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead (optional <space>):<space> or end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=(?:\s)?\:\s|$))?'
            iregx_pattern += r'(?:(?:\s)?\:\s)?'  # remove (optional <space>):<space>
            # group 3 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=$))?$'
            iregx = re.compile(iregx_pattern)

            indicators = iregx.search(indicator)
            if indicators is not None:
                indicator_list = list(indicators.groups())

        return indicator_list

    def add_group(self, group_data: dict, **kwargs) -> Union[dict, object]:
        """Add a group to Batch Job.

        .. code-block:: javascript

            {
                "name": "Example Incident",
                "type": "Incident",
                "attribute": [{
                    "type": "Description",
                    "displayed": false,
                    "value": "Example Description"
                }],
                "xid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904",
                "associatedGroupXid": [
                    "e336e2dd-5dfb-48cd-a33a-f8809e83e904:58",
                ],
                "tag": [{
                    "name": "China"
                }]
            }

        Args:
            group_data: The full Group data including attributes, labels, tags, and
                associations.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Union[dict, object]: The new group dict/object or the previously stored dict/object.
        """
        return self._group(group_data, kwargs.get('store', True))

    def add_indicator(self, indicator_data: dict, **kwargs) -> Union[dict, object]:
        """Add an indicator to Batch Job.

        .. code-block:: javascript

            {
                "type": "File",
                "rating": 5.00,
                "confidence": 50,
                "summary": "53c3609411c83f363e051d455ade78a7
                            : 57a49b478310e4313c54c0fee46e4d70a73dd580
                            : db31cb2a748b7e0046d8c97a32a7eb4efde32a0593e5dbd58e07a3b4ae6bf3d7",
                "associatedGroups": [
                    {
                        "groupXid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904"
                    }
                ],
                "attribute": [{
                    "type": "Source",
                    "displayed": true,
                    "value": "Malware Analysis provided by external AMA."
                }],
                "fileOccurrence": [{
                    "fileName": "drop1.exe",
                    "date": "2017-03-03T18:00:00-06:00"
                }],
                "tag": [{
                    "name": "China"
                }],
                "xid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904:170139"
            }

        Args:
            indicator_data: The Full Indicator data including attributes, labels, tags,
                and associations.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Union[dict, object]: The new group dict/object or the previously stored dict/object.
        """
        if indicator_data.get('type') not in ['Address', 'EmailAddress', 'File', 'Host', 'URL']:
            # for custom indicator types the valueX fields are required.
            # using the summary we can build the values
            index = 1
            for value in self._indicator_values(indicator_data.get('summary')):
                indicator_data[f'value{index}'] = value
                index += 1
        if indicator_data.get('type') == 'File':
            # convert custom field name to the appropriate value for batch v2
            size = indicator_data.pop('size', None)
            if size is not None:
                indicator_data['intValue1'] = size
        if indicator_data.get('type') == 'Host':
            # convert custom field name to the appropriate value for batch v2
            dns_active = indicator_data.pop('dnsActive', None)
            if dns_active is not None:
                indicator_data['flag1'] = dns_active
            whois_active = indicator_data.pop('whoisActive', None)
            if whois_active is not None:
                indicator_data['flag2'] = whois_active
        return self._indicator(indicator_data, kwargs.get('store', True))

    def address(self, ip: str, **kwargs) -> Address:
        """Add Address data to Batch object.

        Args:
            ip: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Address: An instance of the Address class.
        """
        indicator_obj = Address(ip, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def adversary(self, name: str, **kwargs) -> Adversary:
        """Add Adversary data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Adversary: An instance of the Adversary class.
        """
        group_obj = Adversary(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def asn(self, as_number: str, **kwargs) -> ASN:
        """Add ASN data to Batch object.

        Args:
            as_number: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            ASN: An instance of the ASN class.
        """
        indicator_obj = ASN(as_number, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def campaign(self, name: str, **kwargs) -> Campaign:
        """Add Campaign data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            first_seen (str, kwargs): The first seen datetime expression for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Campaign: An instance of the Campaign class.
        """
        group_obj = Campaign(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def cidr(self, block: str, **kwargs) -> CIDR:
        """Add CIDR data to Batch object.

        Args:
            block: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            CIDR: An instance of the CIDR class.
        """
        indicator_obj = CIDR(block, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def close(self) -> None:
        """Cleanup batch job."""
        self.dump()

        # cleanup shelf files
        try:
            self.groups_shelf.close()
            os.unlink(self.group_shelf_fqfn)
        except Exception as ex:
            self.tcex.log.warning(
                f'action=batch-close, filename={self.group_shelf_fqfn} exception={ex}'
            )

        # cleanup shelf files
        try:
            self.indicators_shelf.close()
            os.unlink(self.indicator_shelf_fqfn)
        except Exception as ex:
            self.tcex.log.warning(
                f'action=batch-close, filename={self.indicator_shelf_fqfn} exception={ex}'
            )

    @property
    def data(self):
        """Return the batch indicator/group and file data to be sent to the ThreatConnect API.

        **Processing Order:**
        * Process groups in memory up to max batch size.
        * Process groups in shelf to max batch size.
        * Process indicators in memory up to max batch size.
        * Process indicators in shelf up to max batch size.

        This method will remove the group/indicator from memory and/or shelf.

        Returns:
            dict: A dictionary of group, indicators, and/or file data.
        """
        data = {'file': {}, 'group': [], 'indicator': []}
        tracker = {'count': 0, 'bytes': 0}

        # process group from memory, returning if max values have been reached
        if self.data_groups(data, self.groups, tracker) is True:
            return data

        # process group from shelf file, returning if max values have been reached
        if self.data_groups(data, self.groups_shelf, tracker) is True:
            return data

        # process indicator from memory, returning if max values have been reached
        if self.data_indicators(data, self.indicators, tracker) is True:
            return data

        # process indicator from shelf file, returning if max values have been reached
        if self.data_indicators(data, self.indicators_shelf, tracker) is True:
            return data

        return data

    def data_group_association(self, data: dict, tracker: dict, xid: str) -> None:
        """Return group dict array following all associations.

        The *data* dict is passed by reference to make it easier to update both the group data
        and file data inline versus passing the data all the way back up to the calling methods.

        Args:
            data: The data dict to update with group and file data.
            tracker: A dict containing total count of all entities collected and
                the total size in bytes of all entities collected.
            xid: The xid of the group to retrieve associations.
        """
        xids = deque()
        xids.append(xid)

        while xids:
            xid = xids.popleft()  # remove current xid
            group_data = None

            if xid in self.groups:
                group_data = self.groups.get(xid)
                del self.groups[xid]
            elif xid in self.groups_shelf:
                group_data = self.groups_shelf.get(xid)
                del self.groups_shelf[xid]

            if group_data:
                file_data, group_data = self.data_group_type(group_data)
                data['group'].append(group_data)
                if file_data:
                    data['file'][xid] = file_data

                # update entity trackers
                tracker['count'] += 1
                # tracker['bytes'] += sys.getsizeof(json.dumps(group_data))

                # extend xids with any groups associated with the same object
                xids.extend(group_data.get('associatedGroupXid', []))

    @staticmethod
    def data_group_type(group_data: Union[dict, object]) -> Tuple[dict, dict]:
        """Return dict representation of group data and file data.

        Args:
            group_data: The group data dict or object.

        Returns:
            Tuple[dict, dict]: A tuple containing file_data and group_data.
        """
        file_data = {}
        if isinstance(group_data, dict):
            # process file content
            file_content = group_data.pop('fileContent', None)
            if file_content is not None:
                file_data = {
                    'fileContent': file_content,
                    'fileName': group_data.get('fileName'),
                    'type': group_data.get('type'),
                }
        else:
            # get the file data from the object and return dict format of object
            if group_data.data.get('type') in ['Document', 'Report']:
                file_data = group_data.file_data
            group_data = group_data.data

        return file_data, group_data

    def data_groups(self, data: dict, groups: list, tracker: dict) -> bool:
        """Process Group data.

        Args:
            data: The data dict to update with group and file data.
            groups: The list of groups to process.
            tracker: A dict containing total count of all entities collected and
                the total size in bytes of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # convert groups.keys() to a list to prevent dictionary change error caused by
        # the data_group_association function deleting items from the object.

        # process group objects
        for xid in list(groups.keys()):
            # get association from group data
            self.data_group_association(data, tracker, xid)

            if tracker.get('count') % 10_000 == 0:
                # log count/size at a sane level
                self.tcex.log.debug(
                    '''feature=batch, action=data-groups, '''
                    f'''count={tracker.get('count'):,}, bytes={tracker.get('bytes'):,}'''
                )

        return False

    def data_indicators(self, data: dict, indicators: list, tracker: dict) -> bool:
        """Process Indicator data.

        Args:
            data: The data dict to update with group and file data.
            indicators: The list of indicators to process.
            tracker: A dict containing total count of all entities collected and
                the total size in bytes of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # process indicator objects
        for xid, indicator_data in list(indicators.items()):
            if not isinstance(indicator_data, dict):
                indicator_data = indicator_data.data
            data['indicator'].append(indicator_data)
            del indicators[xid]

            # update entity trackers
            tracker['count'] += 1
            # tracker['bytes'] += sys.getsizeof(json.dumps(indicator_data))

            if tracker.get('count') % 10_000 == 0:
                # log count/size at a sane level
                self.tcex.log.debug(
                    '''feature=batch, action=data-indicators, '''
                    f'''count={tracker.get('count'):,}, bytes={tracker.get('bytes'):,}'''
                )

        return False

    def document(self, name: str, file_name: str, **kwargs) -> Document:
        """Add Document data to Batch object.

        Args:
            name: The name for this Group.
            file_name: The name for the attached file for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            file_content (str;method, kwargs): The file contents or
                callback method to retrieve file content.
            malware (bool, kwargs): If true the file is considered malware.
            password (bool, kwargs): If malware is true a password for the zip archive is
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Document: An instance of the Document class.
        """
        group_obj = Document(name, file_name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def dump(self) -> None:
        """Process Batch request to ThreatConnect API."""
        content = self.data
        content.pop('file', {})
        if not content.get('group') and not content.get('indicator'):
            return

        # special code for debugging App using batchV2.
        self.write_batch_json(content)

        # store the length of the batch data to use for poll interval calculations
        self.tcex.log.info(
            '''feature=batch, event=dump, type=group, ''' f'''count={len(content.get('group')):,}'''
        )
        self.tcex.log.info(
            '''feature=batch, event=dump, type=indicator, '''
            f'''count={len(content.get('indicator')):,}'''
        )
        self.tcex.log.info(f'''feature=batch, event=dump, type=batch, size={self._batch_size:,}''')

        # reset batch size after dump
        self._batch_size = 0

    def email(self, name: str, subject: str, header: str, body: str, **kwargs) -> Email:
        """Add Email data to Batch object.

        Args:
            name: The name for this Group.
            subject: The subject for this Email.
            header: The header for this Email.
            body: The body for this Email.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            from_addr (str, kwargs): The **from** address for this Email.
            to_addr (str, kwargs): The **to** address for this Email.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Email: An instance of the Email class.
        """
        group_obj = Email(name, subject, header, body, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def email_address(self, address: str, **kwargs) -> EmailAddress:
        """Add Email Address data to Batch object.

        Args:
            address: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            EmailAddress: An instance of the EmailAddress class.
        """
        indicator_obj = EmailAddress(address, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def event(self, name: str, **kwargs) -> Event:
        """Add Event data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            event_date (str, kwargs): The event datetime expression for this Group.
            status (str, kwargs): The status for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Event: An instance of the Event class.
        """
        group_obj = Event(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def file(
        self,
        md5: Optional[str] = None,
        sha1: Optional[str] = None,
        sha256: Optional[str] = None,
        **kwargs,
    ) -> File:
        """Add File data to Batch object.

        .. note:: A least one file hash value must be specified.

        Args:
            md5: The md5 value for this Indicator.
            sha1: The sha1 value for this Indicator.
            sha256: The sha256 value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            size (str, kwargs): The file size for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            File: An instance of the File class.

        """
        indicator_obj = File(md5, sha1, sha256, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    @staticmethod
    def generate_xid(identifier: Optional[Union[list, str]] = None):
        """Generate xid from provided identifiers.

        .. Important::  If no identifier is provided a unique xid will be returned, but it will
                        not be reproducible. If a list of identifiers are provided they must be
                        in the same order to generate a reproducible xid.

        Args:
            identifier:  Optional *string* value(s) to be
               used to make a unique and reproducible xid.

        """
        if identifier is None:
            identifier = str(uuid.uuid4())
        elif isinstance(identifier, list):
            identifier = '-'.join([str(i) for i in identifier])
            identifier = hashlib.sha256(identifier.encode('utf-8')).hexdigest()
        return hashlib.sha256(identifier.encode('utf-8')).hexdigest()

    def group(self, group_type: str, name: str, **kwargs) -> object:
        """Add Group data to Batch object.

        Args:
            group_type: The ThreatConnect define Group type.
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            object: An instance of one of the Group classes.
        """
        group_obj = Group(group_type, name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    @property
    def group_shelf_fqfn(self):
        """Return groups shelf fully qualified filename.

        For testing/debugging a previous shelf file can be copied into the tc_temp_path directory
        instead of creating a new shelf file.
        """
        if self._group_shelf_fqfn is None:
            # new shelf file
            self._group_shelf_fqfn = os.path.join(
                self.tcex.args.tc_temp_path, f'groups-{str(uuid.uuid4())}'
            )
        return self._group_shelf_fqfn

    @property
    def groups(self) -> dict:
        """Return dictionary of all Groups data."""
        if self._groups is None:
            # plain dict, but could be something else in future
            self._groups = {}
        return self._groups

    @property
    def groups_shelf(self) -> object:
        """Return dictionary of all Groups data."""
        if self._groups_shelf is None:
            self._groups_shelf = shelve.open(self.group_shelf_fqfn, writeback=False)
        return self._groups_shelf

    def host(self, hostname: str, **kwargs) -> Host:
        """Add Host data to Batch object.

        Args:
            hostname: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            dns_active (bool, kwargs): If True DNS active is enabled for this indicator.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            whois_active (bool, kwargs): If True WhoIs active is enabled for this indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Host: An instance of the Host class.
        """
        indicator_obj = Host(hostname, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def incident(self, name: str, **kwargs) -> Incident:
        """Add Incident data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            event_date (str, kwargs): The event datetime expression for this Group.
            status (str, kwargs): The status for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Incident: An instance of the Incident class.
        """
        group_obj = Incident(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def indicator(self, indicator_type: str, summary: str, **kwargs) -> object:
        """Add Indicator data to Batch object.

        Args:
            indicator_type: The ThreatConnect define Indicator type.
            summary: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            object: An instance of one of the Indicator classes.
        """
        indicator_obj = Indicator(indicator_type, summary, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    @property
    def indicator_shelf_fqfn(self) -> str:
        """Return indicator shelf fully qualified filename.

        For testing/debugging a previous shelf file can be copied into the tc_temp_path directory
        instead of creating a new shelf file.
        """
        if self._indicator_shelf_fqfn is None:
            # new shelf file
            self._indicator_shelf_fqfn = os.path.join(
                self.tcex.args.tc_temp_path, f'indicators-{str(uuid.uuid4())}'
            )
        return self._indicator_shelf_fqfn

    @property
    def indicators(self) -> dict:
        """Return dictionary of all Indicator data."""
        if self._indicators is None:
            # plain dict, but could be something else in future
            self._indicators = {}
        return self._indicators

    @property
    def indicators_shelf(self) -> object:
        """Return dictionary of all Indicator data."""
        if self._indicators_shelf is None:
            self._indicators_shelf = shelve.open(self.indicator_shelf_fqfn, writeback=False)
        return self._indicators_shelf

    def intrusion_set(self, name: str, **kwargs) -> IntrusionSet:
        """Add Intrusion Set data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            IntrusionSet: An instance of the IntrusionSet class.
        """
        group_obj = IntrusionSet(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def mutex(self, mutex: str, **kwargs) -> Mutex:
        """Add Mutex data to Batch object.

        Args:
            mutex: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Mutex: An instance of the Mutex class.
        """
        indicator_obj = Mutex(mutex, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def registry_key(
        self, key_name: str, value_name: str, value_type: str, **kwargs
    ) -> RegistryKey:
        """Add Registry Key data to Batch object.

        Args:
            key_name: The key_name value for this Indicator.
            value_name: The value_name value for this Indicator.
            value_type: The value_type value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            RegistryKey: An instance of the Registry Key class.
        """
        indicator_obj = RegistryKey(key_name, value_name, value_type, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def report(self, name: str, **kwargs) -> Report:
        """Add Report data to Batch object.

        Args:
            name: The name for this Group.
            file_name (str): The name for the attached file for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            file_content (str;method, kwargs): The file contents or callback method to retrieve
                file content.
            publish_date (str, kwargs): The publish datetime expression for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Report: An instance of the Report class.
        """
        group_obj = Report(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def save(self, resource: Union[dict, object]) -> None:
        """Save group|indicator dict or object to shelve.

        Best effort to save group/indicator data to disk.  If for any reason the save fails
        the data will still be accessible from list in memory.

        Args:
            resource: The Group or Indicator dict or object.
        """
        resource_type = None
        xid = None
        if isinstance(resource, dict):
            resource_type = resource.get('type')
            xid = resource.get('xid')
        else:
            resource_type = resource.type
            xid = resource.xid

        if resource_type is not None and xid is not None:
            saved = True

            if resource_type in self.tcex.group_types:
                try:
                    # groups
                    self.groups_shelf[xid] = resource
                except Exception:
                    saved = False

                if saved:
                    try:
                        del self._groups[xid]
                    except KeyError:
                        # if group was saved twice it would already be delete
                        pass
            elif resource_type in self.tcex.indicator_types_data.keys():
                try:
                    # indicators
                    self.indicators_shelf[xid] = resource
                except Exception:
                    saved = False

                if saved:
                    try:
                        del self._indicators[xid]
                    except KeyError:
                        # if indicator was saved twice it would already be delete
                        pass

    def signature(
        self, name: str, file_name: str, file_type: str, file_text: str, **kwargs
    ) -> Signature:
        """Add Signature data to Batch object.

        Valid file_types:
        + Snort ®
        + Suricata
        + YARA
        + ClamAV ®
        + OpenIOC
        + CybOX ™
        + Bro
        + Regex
        + SPL - Splunk ® Search Processing Language

        Args:
            name: The name for this Group.
            file_name: The name for the attached signature for this Group.
            file_type: The signature type for this Group.
            file_text: The signature content for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Signature: An instance of the Signature class.
        """
        group_obj = Signature(name, file_name, file_type, file_text, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def threat(self, name: str, **kwargs) -> Threat:
        """Add Threat data to Batch object

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Threat: An instance of the Threat class.
        """
        group_obj = Threat(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def user_agent(self, text: str, **kwargs) -> UserAgent:
        """Add User Agent data to Batch object

        Args:
            text: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            UserAgent: An instance of the UserAgent class.
        """
        indicator_obj = UserAgent(text, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def url(self, text: str, **kwargs) -> URL:
        """Add URL Address data to Batch object.

        Args:
            text (str): The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            URL: An instance of the URL class.
        """
        indicator_obj = URL(text, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def write_batch_json(self, content: dict) -> None:
        """Write batch json data to a file."""
        if content:
            # get timestamp as a string without decimal place and consistent length
            filename = f'{str(round(time.time() * 10000000))}.json.gz'
            if self.output_extension is not None:
                # add any additional extension provided
                filename += self.output_extension
            # TODO: is this needed
            self._batch_files.append(filename)
            fqfn = os.path.join(self.output_dir, filename)
            with gzip.open(fqfn, mode='wt', encoding='utf-8') as fh:
                json.dump(content, fh)

            # send callback the filename
            if callable(self.write_callback):
                self.write_callback(fqfn, **self.write_callback_kwargs)



#!/usr/bin/env python

"""Lark based STIX2 Pattern Parser"""

# standard library
import os

# third-party
from lark import Lark, Transformer, v_args


class Indicator:
    """Indicator Object as recognized by the parser.

    Indicators have path and value properties, and may be
    accessed as if they are dictionaries, e.g. indicator['path']
    is the same as indicator.path, for compatibility with an
    older STIX parser.
    """

    def __init__(self, path, value):
        """Init"""

        self.path = path
        self.value = value

    def __repr__(self):
        """Repr"""

        return f'<Indicator({self.path!r}) = {self.value!r}>'

    def __str__(self):
        """Str"""

        return self.value

    def __getitem__(self, key, default=None):
        """__getitem__"""

        return getattr(self, key, default)


@v_args(inline=True)
class Stix2IndicatorTransformer(Transformer):
    """Walks the parse tree and evaluates the results"""

    # N.B.  It may be necessary to define more productions
    # if, for example, complex rules end up including junk
    # in the parse.

    # Right now, the indciator and indicator_list productions
    # generate a single and list of Indicator objects.  The
    # observation_expression_and and observation_expression_or
    # productions propagate those indicators upwards in the tree.

    # The default Lark production is to create Lark.Tree(production, children, meta)
    # where meta is metadata about terminals and is unused.

    @staticmethod
    def extract_indicators(tree):
        """Pull out the indicators from this part of the parse tree"""

        indicators = []

        if isinstance(tree, list):  # a list MUST be a list of indicators
            indicators = tree
        elif isinstance(tree, Indicator):
            indicators = [tree]
        else:  # otherwise this is a Tree node, so iterate through it
            for subtree in tree.iter_subtrees():
                for child in subtree.children:
                    if not isinstance(child, list):
                        child = [child]
                        # child nodes can be Tree nodes or indicators, but
                        # iter_subtrees will find the trees, so we only care
                        # about the Indicator children
                    for subchild in child:
                        if isinstance(subchild, Indicator):
                            indicators.append(subchild)

        return indicators

    pattern = extract_indicators  # pattern productions = all indicators

    @staticmethod
    def object_path(*args):
        """object_path production"""

        result = args[0] + ':' + args[1]
        if len(args) > 2:
            result += '.'.join(args[2:])
        return result

    @staticmethod
    def path_step(a, b):
        """path_step production"""

        return f'{a} {b}'

    @staticmethod
    def key_path_step(a):
        """key_path_step production"""

        return f'.{a}'

    @staticmethod
    def index_path_step(a):
        """index_path_step production"""

        return f'[{a}]'

    @staticmethod
    def indicator(*args):
        """Do indicator production"""

        path = args[0]
        value = args[-1]
        return Indicator(path, value.value[1:-1])

    @staticmethod
    def set_literal(*args):
        """Do set_literal production"""

        result = []
        for arg in args:
            if isinstance(arg, list):
                result.append(arg)
            else:
                result.append(str(arg)[1:-1])
        return result

    @staticmethod
    def indicator_list(path, values):
        """Do indicator_list production"""

        indicators = [Indicator(path, value) for value in values]
        return indicators

    @staticmethod
    def chain_left_right(left, right):
        """Chain together the left and right"""

        # Glue together the left & right hand lists

        if not right:
            return left

        if not isinstance(left, list):
            left = [
                left,
            ]

        if not isinstance(right, list):
            right = [
                right,
            ]

        left.extend(right)
        return left

    observation_expression_or = chain_left_right
    observation_expression_and = chain_left_right


class Stix2IndicatorParser:
    """STIX2 Parser using Lark.

    This parser only extracts positive indicators from the STIX2 pattern field.
    """

    def __init__(self):
        """STIX2 Parser Initialization"""

        self.result = None
        grammar = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'stix2.lark')
        self.parser = Lark.open(
            grammar, parser='lalr', start='pattern', transformer=Stix2IndicatorTransformer()
        )

    def parse(self, stix2pattern):
        """Parse Stix2 Pattern, pulling out indicators.

        Returns a list of Indicator objects
        """

        result = self.parser.parse(stix2pattern)
        self.result = result

        return result



"""ThreatConnect TI Threat"""
from ..group import Group


class Threat(Group):
    """Unique API calls for Threat API Endpoints"""

    def __init__(self, ti: 'ThreatIntelligenc', **kwargs):
        """Initialize Class Properties.

        Args:
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
        """
        super().__init__(ti, sub_type='Threat', api_entity='threat', api_branch='threats', **kwargs)
